
//
//  ase.cpp - *.ASE file parser
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This is the ASE parser.
//
//  NOTE: Even though MAX uses XYZ, this uses XYZW and sets W to 0 to keep 128bit aligned!
//      Also, heavily overloaded with Assertions so any future expansion that breaks the
//      recursion and pointer validity, will be identified quickly.
//
//  Created - 13 Mar, 2002 - James Leiterman
//
//  Last Modification
//
//  stab=4
//

#include    "KariType.h"                                // Master Definitions

    //  3rd Party Files

#include    <iostream.h>
#include    <assert.h>
#include    <stdio.h>
#include    <string.h>
#include    <stdlib.h>
#include    <math.h>

    //  Book Development Library Modules
    
#include    "util.h"                                    // General purpose utilities
#include    "vmp3D.h"
#include    "ase.h"                                     // ASE file parser
#include    "aseEx.h"                                   // ASE file parser (private data)

#include    "ram.h"                                     // Memory Module




//
//  Key table Lookup
//
//  NOTE: Keep in ENUM orderand Alphabetical order!
//

AseLabel AseKeyTbl[] = {
    { iKEY_UNKNOWN, "" },
        //  Insert new enums below this line!
    { iKEY_BITMAP,              xKEY_BITMAP },
    { iKEY_GEOMOBJECT,          xKEY_GEOMOBJECT },
    { iKEY_MAP_DIFFUSE,         xKEY_MAP_DIFFUSE },
    { iKEY_MAP_NAME,            xKEY_MAP_NAME },
    { iKEY_MAP_OPACITY,         xKEY_MAP_OPACITY },
    { iKEY_MATERIAL,            xKEY_MATERIAL },
    { iKEY_MATERIAL_COUNT,      xKEY_MATERIAL_COUNT },
    { iKEY_MATERIAL_LIST,       xKEY_MATERIAL_LIST },
    { iKEY_MATERIAL_NAME,       xKEY_MATERIAL_NAME },
    { iKEY_MESH,                xKEY_MESH },
    { iKEY_MESH_ANIMATION,      xKEY_MESH_ANIMATION },
    { iKEY_MESH_CFACE,          xKEY_MESH_CFACE },
    { iKEY_MESH_CFACELIST,      xKEY_MESH_CFACELIST },
    { iKEY_MESH_CVERTLIST,      xKEY_MESH_CVERTLIST },
    { iKEY_MESH_FACE,           xKEY_MESH_FACE },
    { iKEY_MESH_FACENORMAL,     xKEY_MESH_FACENORMAL },
    { iKEY_MESH_FACE_LIST,      xKEY_MESH_FACE_LIST },
    { iKEY_MESH_MTLID,          xKEY_MESH_MTLID },
    { iKEY_MESH_NORMALS,        xKEY_MESH_NORMALS },
    { iKEY_MESH_NUMCVERTEX,     xKEY_MESH_NUMCVERTEX },
    { iKEY_MESH_NUMCVFACES,     xKEY_MESH_NUMCVFACES },
    { iKEY_MESH_NUMFACES,       xKEY_MESH_NUMFACES },
    { iKEY_MESH_NUMTVERTEX,     xKEY_MESH_NUMTVERTEX },
    { iKEY_MESH_NUMTVFACES,     xKEY_MESH_NUMTVFACES },
    { iKEY_MESH_NUMVERTEX,      xKEY_MESH_NUMVERTEX },
    { iKEY_MESH_SMOOTHING,      xKEY_MESH_SMOOTHING },
    { iKEY_MESH_TFACE,          xKEY_MESH_TFACE },
    { iKEY_MESH_TFACELIST,      xKEY_MESH_TFACELIST },
    { iKEY_MESH_TVERT,          xKEY_MESH_TVERT },
    { iKEY_MESH_TVERTLIST,      xKEY_MESH_TVERTLIST },
    { iKEY_MESH_VERTCOL,        xKEY_MESH_VERTCOL },
    { iKEY_MESH_VERTEX,         xKEY_MESH_VERTEX },
    { iKEY_MESH_VERTEXNORMAL,   xKEY_MESH_VERTEXNORMAL },
    { iKEY_MESH_VERTEX_LIST,    xKEY_MESH_VERTEX_LIST },
    { iKEY_NODE_NAME,           xKEY_NODE_NAME },
    { iKEY_NODE_TM,             xKEY_NODE_TM },
    { iKEY_NUMSUBMTLS,          xKEY_NUMSUBMTLS },
    { iKEY_SCENE,               xKEY_SCENE },
    { iKEY_SCENE_FIRSTFRAME,    xKEY_SCENE_FIRSTFRAME },
    { iKEY_SCENE_FRAMESPEED,    xKEY_SCENE_FRAMESPEED },
    { iKEY_SCENE_LASTFRAME,     xKEY_SCENE_LASTFRAME },
    { iKEY_SCENE_TICKSPERFRAME, xKEY_SCENE_TICKSPERFRAME },
    { iKEY_SHAPEOBJECT,         xKEY_SHAPEOBJECT },
    { iKEY_SHAPE_LINE,          xKEY_SHAPE_LINE },
    { iKEY_SHAPE_LINECOUNT,     xKEY_SHAPE_LINECOUNT },
    { iKEY_SHAPE_VERTEXCOUNT,   xKEY_SHAPE_VERTEXCOUNT },
    { iKEY_SHAPE_VERTEX_KNOT,   xKEY_SHAPE_VERTEX_KNOT },
    { iKEY_SUBMATERIAL,         xKEY_SUBMATERIAL },
    { iKEY_TM_POS,              xKEY_TM_POS },
    { iKEY_TM_ROTANGLE,         xKEY_TM_ROTANGLE },
    { iKEY_TM_ROTAXIS,          xKEY_TM_ROTAXIS },
    { iKEY_TM_SCALE,            xKEY_TM_SCALE },
    { iKEY_TM_SCALEAXIS,        xKEY_TM_SCALEAXIS },
    { iKEY_TM_SCALEAXISANG,     xKEY_TM_SCALEAXISANG },

};
#define KeyTblCnt   (sizeof(AseKeyTbl) / sizeof(AseLabel))



//
//  ASE Integrity Checks
//
#ifdef ASSERT
void aseIntegrityCheck( const char * pStr, ulong nLen )
{
    uint n;
    AseLabel *pRec;

    ASSERT_PTR(pStr);
    ASSERT(iKEY_MAX == KeyTblCnt);
    ASSERT(0== *(pStr+nLen));                           // Verify a terminator exists!

    pRec = AseKeyTbl;
    n = 0;
    do {
        ASSERT(n==(uint)(pRec->keyId));                 // Enum is the index?
        ASSERT(pRec->pszName);                          // Name exists?

        if (iKEY_MAX <= ++n)
          {
            break;                                      // Break before testing one too many names!
          }

            //  Is it properly alpha sorted???
        ASSERT(0 > strcmp(pRec->pszName,(pRec+1)->pszName));
        
        pRec++;
      } while(1);

}
#endif





//
//  Get Face Information
//
//  NOTE: sscanf is too slow so did parsing manually.
//
//  Always A: # B: # C: # AB: # BC: # CA: # sequenced!
//
static char *getFace( Face *pFace, const char *p )
{
    ASSERT_PTR(pFace);
    ASSERT_PTR(p);

    p = skipspace(p);
    p = skipspace(skipword(p));     // A:
    pFace->a = atoi(p);

    p = skipspace(skipword(p));     // #
    p = skipspace(skipword(p));     // B:
    pFace->b = atoi(p);

    p = skipspace(skipword(p));     // #
    p = skipspace(skipword(p));     // C:
    pFace->c = atoi(p);

    p = skipspace(skipword(p));     // #
    p = skipspace(skipword(p));     // AB:
    pFace->ab = atoi(p);

    p = skipspace(skipword(p));     // #
    p = skipspace(skipword(p));     // BC:
    pFace->bc = atoi(p);

    p = skipspace(skipword(p));     // #
    p = skipspace(skipword(p));     // CA:
    pFace->ca = atoi(p);

    return (char *) skipword(p);    // #
}



//
//  Get TFace & CFace Information
//
//  # # #
//
static char *getTFace( TFace *pTFace, const char *p )
{
    ASSERT_PTR(pTFace);
    ASSERT_PTR(p);

    p = skipspace(p);   
    pTFace->a = atoi(p);

    p = skipspace(skipword(p)); 
    pTFace->b = atoi(p);

    p = skipspace(skipword(p)); 
    pTFace->c = atoi(p);

    return (char *) skipword(p);
}



//
//  Get 3D Vector
//
//  #f #f #f
//

static char *getVert( vmp3DQDVector *pVert, const char *p )
{
    ASSERT_PTR(pVert);
    ASSERT_PTR(p);

    pVert->x = pASCIItoDouble( &p );
    pVert->y = pASCIItoDouble( &p );
    pVert->z = pASCIItoDouble( &p );

    return (char *)p;
}



//
//  Get label KeyId
//
static bool aseGetLabelId( const char *pName, iKEY* pnKey)
{
    int i, idx, lo, hi;

    ASSERT_PTR(pName);
    ASSERT_PTR(pnKey);

        //  Loop until match or not found!

    idx = iKEY_MAX / 2;                         // Start in middle
    lo = iKEY_UNKNOWN+1;
    hi = iKEY_MAX-1;

    do {
        i = strcmp(pName, AseKeyTbl[idx].pszName);      // Compare Labels
        if (0==i)                                       // Match found?
          {
            *pnKey = (iKEY)idx;                         // return KeyId
            return true;                                // ***Success***
          }

        else if (i<0)                                   // Left of index
          {
            hi = idx-1;
          }

        else                                            // right of index
          {
            lo = idx+1;
          }

        idx = (hi + lo) >> 1;                           // Set new mid point for index
      } while(lo<=hi);

    return false;                                       // Not Found!
}



//
//  Node creation append
//

static bool NodeCreate( Node *pNodeList, Node **ppNewNode )
{
    Node *pNew;

        //  Allocate a Node

    if (!ramGetClr((byte**)&pNew, sizeof(Node)))
      {
        return false;
      }

    pNew->keyId = iKEY_UNKNOWN;

    pNew->next = pNodeList->next;
    pNodeList->next = pNew;                             // Insert link

    *ppNewNode = pNew;
    return true;
}



//
//  ASE parser nest
//

static bool aseParseNest( const char ** ppStr, iKEY keyId, Node * pNodeBase, MeshArgs *ma, rdbInfo *pStat )
{
    char    field[256];
    const char *p;
    bool    ret;
    Node    *pNode;
    char    *pcTmp;
    vmp3DQDVector *pVert;
    uint    nSlot, nCnt, nVSlot;
    Face    *pFace;
    Normal  *pNormal;

    pVert = NULL;
    nVSlot = 0;

    ASSERT_PTR(ppStr);
    ASSERT(keyId <= iKEY_MAX);

    ret = true;

        //  If pNodeBase is NULL then we ignore all data from this level or deeper!
        //  We just use the parser to follow the '{'  '}' nesting.

    pNode = pNodeBase;                                  // Node from parent is current node!
    p = *ppStr;

        //  Loop for lines
        //  NOTE: We come in knowing our parent's iKEY!

    while (*p)
      {
        p = getname( field, skipspace(p) );

            //  If we care about this data type, and it's a label...

        if ((NULL!=pNode) && ('*' == *field))           // Label?
          {
            if (!aseGetLabelId( field+1, &keyId))       // One we care about?
              {
                keyId = iKEY_UNKNOWN;                   // Unknown iKEY type
                p = skipline(p);                        // Skip rest of line or until '{'
              }                                         // If '{' we recursively skip!
            else
              {
                void *pData;

                pData = NULL;

                switch(keyId)
                  {
                case iKEY_SCENE:                        // {  Level 0 Header
                    if (!ramGetClr((byte**)&pData, sizeof(NodeScene)) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    ma->pScene = (NodeScene*)pData;
                    goto pBuild;                        // Allocate a Node

                case iKEY_SCENE_FIRSTFRAME:
                    ASSERT_PTR(pNode->pData);
                    ASSERT(iKEY_SCENE==pNode->keyId);

                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );
                    ma->pScene->nFirstFrame = nSlot;
                    ma->nCurrentFrame = 0;
                    break;

                case iKEY_SCENE_LASTFRAME:
                    ASSERT_PTR(pNode->pData);
                    ASSERT(iKEY_SCENE==pNode->keyId);

                    p = getword( field, skipspace(p) );
                    ma->pScene->nLastFrame = atoi( field );
                    break;

                case iKEY_SCENE_FRAMESPEED:
                    ASSERT_PTR(pNode->pData);
                    ASSERT(iKEY_SCENE==pNode->keyId);

                    p = getword( field, skipspace(p) );
                    ma->pScene->nFrameSpeed = atoi( field );
                    break;

                case iKEY_SCENE_TICKSPERFRAME:
                    ASSERT_PTR(pNode->pData);
                    ASSERT(iKEY_SCENE==pNode->keyId);

                    p = getword( field, skipspace(p) );
                    ma->pScene->TicksPerFrame = atoi( field );
                    break;

                case iKEY_GEOMOBJECT:
                    if (!ramGetClr((byte**)&pData, sizeof(GeomObject)) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    nCnt = ma->pScene->nLastFrame - ma->pScene->nFirstFrame;
                    ASSERT(0 < (int)nCnt);

                    ma->pGeomObject = (GeomObject*)pData;
                    ma->pGeomObject->nAnimCnt = nCnt;

                    if (!ramGetClr((byte**)&(ma->pGeomObject->pAni), sizeof(Mesh)*(nCnt+2)) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    goto pBuild;                        // Allocate a Node

                case iKEY_NODE_TM:
                    break;

                case iKEY_TM_POS:           // # # #
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((ShapeObject*)(pNode->pData))->vPos), p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((GeomObject*)(pNode->pData))->vPos), p); 
                      }
                    break;

                case iKEY_TM_ROTAXIS:       // # # #
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((ShapeObject*)(pNode->pData))->vRotAxis), p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((GeomObject*)(pNode->pData))->vRotAxis), p); 
                      }
                    break;

                case iKEY_TM_ROTANGLE:      // # 
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        ((ShapeObject*)(pNode->pData))->RotAngle = pASCIItoDouble(&p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        ((GeomObject*)(pNode->pData))->RotAngle = pASCIItoDouble(&p); 
                      }
                    break;

                case iKEY_TM_SCALE:         // # # #
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((ShapeObject*)(pNode->pData))->vScale), p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((GeomObject*)(pNode->pData))->vScale), p); 
                      }
                    break;

                case iKEY_TM_SCALEAXIS:     // # # #
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((ShapeObject*)(pNode->pData))->vScaleAxis), p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        p = getVert(&(((GeomObject*)(pNode->pData))->vScaleAxis), p); 
                      }
                    break;

                case iKEY_TM_SCALEAXISANG:  // #
                    if (iKEY_SHAPEOBJECT == pNode->keyId)
                      {
                        ((ShapeObject*)(pNode->pData))->ScaleAxisAng = pASCIItoDouble(&p); 
                      }
                    else if (iKEY_GEOMOBJECT == pNode->keyId)
                      {
                        ((GeomObject*)(pNode->pData))->ScaleAxisAng = pASCIItoDouble(&p); 
                      }
                    break;

                case iKEY_MATERIAL_LIST:                // {  Level 0 Header
                    ma->pGeomObject = NULL;
pBuild:
                    ma->pMaterial = NULL;
                    ma->pSubMaterial = NULL;
                    ma->pShape = NULL;
                    ma->pShapeLine = NULL;
                    ma->pMesh = NULL;
                    ma->nCurrentFrame = 0;

                    if (!NodeCreate(pNodeBase, &pNode))
                      {
                        ret = false;
                        break;
                      }

                    pNode->pData = pData;               // Assign data structure
                    pNode->keyId = keyId;
                    pNodeBase = pNode;                  // New Base
                    break;

                case iKEY_MATERIAL_COUNT:               // #
                    p = getword( field, skipspace(p) );
                    pNode->nItemCnt = atoi( field );

                    if (pNode->nItemCnt)
                      {
                        if (!ramGetClr((byte**)&(pNode->pData), sizeof(NodeMaterial)*pNode->nItemCnt))
                          {
                            ret = false;
                            goto pexit;
                          }

                        ma->pMaterial = ((NodeMaterial*)(pNode->pData ));
                      }
                    break;

                case iKEY_MATERIAL:                     // # (
                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );

                    ASSERT_PTR(ma->pMaterial);
                    ASSERT(ma->pMaterial==pNode->pData);
                    ma->pMaterial->nSlot = nSlot;
                    break;

                case iKEY_MAP_OPACITY:
                    break;

                case iKEY_MATERIAL_NAME:
                    p = getname( field, skipspace(p) );

                    if (!ramGetClr((byte**)&pcTmp, strlen(field)+1) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    strcpy(pcTmp, field);

                    ma->pMaterial = &((NodeMaterial*)(pNode->pData ))[pNode->nSlot];

                    if (NULL==ma->pMaterial->pSubMaterial)  // If no allocated submaterial array...
                      {
                        ma->pMaterial->pName = pcTmp;
                      }
                    else
                      {
                        ma->pSubMaterial = &ma->pMaterial->pSubMaterial[ma->pMaterial->nSlot];
                        ma->pSubMaterial->pName = pcTmp;
                      }
                        
                    break;

                case iKEY_NUMSUBMTLS:                   // #
                    ma->pMaterial = &((NodeMaterial*)(pNode->pData ))[pNode->nSlot];
                    
                    p = getword( field, skipspace(p) );
                    ma->pMaterial->nItemCnt = atoi( field );

                    if (!ramGetClr((byte**)&(ma->pMaterial->pSubMaterial),
                             sizeof(SubMaterial)*ma->pMaterial->nItemCnt) )
                      {
                        ret = false;
                        goto pexit;
                      }
                    break;

                case iKEY_SUBMATERIAL:                  // #
                    ma->pMaterial = &((NodeMaterial*)(pNode->pData ))[pNode->nSlot];

                    p = getword( field, skipspace(p) );
                    ma->pMaterial->nSlot = atoi( field );
                    break;

                case iKEY_MAP_DIFFUSE:
                    break;

                case iKEY_MAP_NAME:
                    p = getname( field, skipspace(p) );

                    if (!ramGetClr((byte**)&pcTmp, strlen(field)+1) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    strcpy(pcTmp, field);
                    ma->pMaterial->pSubMaterial[ ma->pMaterial->nSlot ].pMapName = pcTmp;
                    break;

                case iKEY_BITMAP:
                    p = getname( field, skipspace(p) );

                    if (!ramGetClr((byte**)&pcTmp, strlen(field)+1) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    strcpy(pcTmp, field);
                    ma->pMaterial->pSubMaterial[ ma->pMaterial->nSlot ].pFPath = pcTmp;
                    break;

                case iKEY_MESH_ANIMATION:
                    break;

                case iKEY_MESH:
                    ma->pGeomObject = (GeomObject*)(pNode->pData);
                    
                    ASSERT_PTR(ma->pGeomObject);
//                  ASSERT(ma->pGeomObject==pNode->pData);
                    ASSERT_PTR(ma->pGeomObject->pAni);

                    ma->pGeomObject->pMesh = ma->pGeomObject->pAni + ma->nCurrentFrame;
                    ma->nCurrentFrame++;                // reference
                    ma->pGeomObject->nItemCnt++;
                    break;
                    
                case iKEY_MESH_NUMVERTEX:
                    ASSERT_PTR(pNode->pData);
                    ma->pGeomObject = (GeomObject*)(pNode->pData);

                    ASSERT_PTR(ma->pGeomObject->pMesh);
                    ma->pMesh = ma->pGeomObject->pMesh;

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumVertex = atoi( field );

                    if (!ramGetClr((byte**)&(ma->pMesh->pVertList), sizeof(vmp3DQDVector)* ma->pMesh->NumVertex))
                      {
                        ma->pMesh->NumVertex = 0;
                        ret = false;
                        goto pexit;
                      }

                    ma->nCnt = ma->pMesh->NumVertex;
                    ma->pVertList = ma->pMesh->pVertList;
                    break;

                case iKEY_MESH_NUMFACES:
                    ASSERT_PTR(pNode->pData);
                    ma->pGeomObject = (GeomObject*)(pNode->pData);

                    ASSERT_PTR(ma->pGeomObject->pMesh);
                    ma->pMesh = ma->pGeomObject->pMesh;

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumFaces = atoi( field );

                    if (!ramGetClr((byte**)&(ma->pMesh->pNormalList), sizeof(Normal)* ma->pMesh->NumFaces))
                      {
                        ma->pMesh->NumFaces = 0;
                        ret = false;
                        goto pexit;
                      }

                    if (!ramGetClr((byte**)&(ma->pMesh->pFaceList), sizeof(Face)* ma->pMesh->NumFaces))
                      {
                        ma->pMesh->NumFaces = 0;
                        ret = false;
                        goto pexit;
                      }

                    break;

                case iKEY_MESH_FACE:
                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );

                    ASSERT(ma->pGeomObject==(GeomObject*)(pNode->pData));
                    ASSERT_PTR(ma->pGeomObject);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);
                    ASSERT_PTR(ma->pMesh);
                    ASSERT(nSlot<ma->pMesh->NumFaces);
                    ASSERT_PTR(ma->pMesh->pFaceList);

                    pFace = &ma->pMesh->pFaceList[nSlot];
                    p = getFace(pFace, p );
                    break;

                case iKEY_MESH_SMOOTHING:
                    ASSERT_PTR(pFace);

                    pFace->Smoothing = 0;
                    p = skipspace(p);
                    if (('0' <=*p) && (*p<='9'))
                      {
                        p = getword( field, p );
                        pFace->Smoothing = atoi( field );
                      }
                    break;

                case iKEY_MESH_MTLID:
                    ASSERT_PTR(pFace);
                    p = getword( field, skipspace(p) );
                    pFace->MaterialId = atoi( field );
                    break;

                case iKEY_MESH_NUMTVERTEX:
                    ASSERT_PTR(ma->pMesh);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumTVertex = atoi( field );

                        //  If no TVertex then skip them!

                    if (0==ma->pMesh->NumTVertex)
                      {
                        break;
                      }

                    if (!ramGetClr((byte**)&(ma->pMesh->pTVertList), sizeof(vmp3DQDVector)* ma->pMesh->NumTVertex))
                      {
                        ma->pMesh->NumTVertex = 0;
                        ret = false;
                        goto pexit;
                      }

                    ma->nCnt = ma->pMesh->NumTVertex;
                    ma->pVertList = ma->pMesh->pTVertList;
                    break;

                case iKEY_MESH_NUMTVFACES:
                    ASSERT_PTR(ma->pMesh);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumTFaces = atoi( field );

                    if (!ramGetClr((byte**)&(ma->pMesh->pTFaceList), sizeof(TFace)* ma->pMesh->NumTFaces))
                      {
                        ma->pMesh->NumTFaces = 0;
                        ret = false;
                        goto pexit;
                      }

                    ma->nCnt = ma->pMesh->NumTFaces;
                    ma->pTFaceList = ma->pMesh->pTFaceList;
                    break;

                case iKEY_MESH_NUMCVERTEX:
                    ASSERT_PTR(ma->pMesh);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumCVertex = atoi( field );

                        //  If no TVertex then skip them!

                    if (0==ma->pMesh->NumCVertex)
                      {
                        break;
                      }

                    if (!ramGetClr((byte**)&(ma->pMesh->pCVertList), sizeof(vmp3DQDVector)* ma->pMesh->NumCVertex))
                      {
                        ma->pMesh->NumCVertex = 0;
                        ret = false;
                        goto pexit;
                      }

                    ma->nCnt = ma->pMesh->NumCVertex;
                    ma->pVertList = ma->pMesh->pCVertList;
                    break;

                case iKEY_MESH_NUMCVFACES:
                    ASSERT_PTR(ma->pMesh);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);

                    p = getword( field, skipspace(p) );
                    ma->pMesh->NumCFaces = atoi( field );

                    if (!ramGetClr((byte**)&(ma->pMesh->pCFaceList), sizeof(TFace)* ma->pMesh->NumCFaces))
                      {
                        ma->pMesh->NumCFaces = 0;
                        ret = false;
                        goto pexit;
                      }

                    ma->nCnt = ma->pMesh->NumCFaces;
                    ma->pTFaceList = ma->pMesh->pCFaceList;
                    break;

                case iKEY_MESH_NORMALS:
                    break;

                    //  Normal

                case iKEY_MESH_FACENORMAL:
                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );

                    ASSERT_PTR(ma->pMesh);
                    ASSERT(ma->pMesh==ma->pGeomObject->pMesh);
                    ASSERT(nSlot<ma->pMesh->NumFaces);  // 1 normal every vertex
                    ASSERT_PTR(ma->pMesh->pNormalList);

                    pNormal = ma->pMesh->pNormalList + nSlot;
                    
                    p = getVert(&(pNormal->vFace), p);  // Get vmp3DQDVector
                    nVSlot = 0;
                    break;

                    //  Get Vertices 0...2 for the current Normal

                case iKEY_MESH_VERTEXNORMAL:
                    ASSERT_PTR(pNormal);
                    ASSERT(nVSlot < 3);                     // A,B,C Face Vertices

                    p = getword( field, skipspace(p) );
                    pNormal->idx[nVSlot] = atoi( field );   // Face Index

                    p = getVert(&(pNormal->v[nVSlot++]), p); // Get vmp3DQDVector
                    break;


                case iKEY_MESH_VERTEX_LIST:
                case iKEY_MESH_FACE_LIST:

                case iKEY_MESH_TFACELIST:
                case iKEY_MESH_TVERTLIST:

                case iKEY_MESH_CVERTLIST:
                case iKEY_MESH_CFACELIST:
                    break;

                    //  Retrieve an element of the Vector table referenced by '.pVertBase' of Maximum count '.nCnt'

                case iKEY_MESH_VERTEX:
                case iKEY_MESH_TVERT:
                case iKEY_MESH_VERTCOL:
                case iKEY_SHAPE_VERTEX_KNOT:

                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );

                    ASSERT(nSlot<ma->nCnt);
                    ASSERT_PTR(ma->pVertList);

                    p = getVert(ma->pVertList + nSlot, p);  // Get vmp3DQDVector
                    break;

                case iKEY_MESH_TFACE:
                case iKEY_MESH_CFACE:
                    p = getword( field, skipspace(p) );
                    nSlot = atoi( field );

                    ASSERT(nSlot<ma->nCnt);
                    ASSERT_PTR(ma->pTFaceList);

                    p = getTFace(ma->pTFaceList + nSlot, p);    // Get a TFace, CFace
                    break;




                    //  I personally like to use Line Shapes for path following.

                case iKEY_SHAPEOBJECT:
                    if (!ramGetClr((byte**)&pData, sizeof(ShapeObject)) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    goto pBuild;                        // Allocate a Node

                case iKEY_NODE_NAME:
                    p = getname( field, skipspace(p) );

                    if (!ramGetClr((byte**)&pcTmp, strlen(field)+1) )
                      {
                        ret = false;
                        goto pexit;
                      }

                    strcpy(pcTmp, field);

                    pNode->pName = pcTmp;
                    break;

                case iKEY_SHAPE_LINECOUNT:
                    ma->pShape = (ShapeObject*)(pNode->pData);
                    p = getword( field, skipspace(p) );
                    ma->pShape->nLineCnt = atoi( field );

                    if (!ramGetClr((byte **)&(ma->pShape->pLine),
                            sizeof(ShapeLine*)*ma->pShape->nLineCnt ))
                      {
                        ret = false;
                        goto pexit;
                      }

                    break;

                    //  Set current Shape: Line for the current ShapeObject

                case iKEY_SHAPE_LINE:
                    ma->pShape = (ShapeObject*)(pNode->pData);

                    p = getword( field, skipspace(p) );
                    ma->pShape->nSlot = atoi( field );
                    break;

                    //  Set # of Vertices used by this Line# for this ShapeObject
                    //  and allocate the storage!

                case iKEY_SHAPE_VERTEXCOUNT:
                    p = getword( field, skipspace(p) );
                    nCnt = atoi( field );

                    if (nCnt)                           // # of vertices this line this shape
                      {
                        if (!ramGetClr((byte**)&ma->pShapeLine,
                                 sizeof(ma->pShapeLine)+(sizeof(vmp3DQDVector)*(nCnt-1)) ))
                          {
                            ret = false;
                            goto pexit;
                          }

                        ASSERT(ma->pShape==pNode->pData);
                        ASSERT_PTR(ma->pShape);
                        ASSERT_PTR(ma->pShape->pLine);

                        ma->pShapeLine->nVertexCnt = nCnt;
                        ma->pShape->pLine[ma->pShape->nSlot] = ma->pShapeLine;  // Save empty list

                        ma->nCnt = nCnt;
                        ma->pVertList = ma->pShapeLine->VList;

                        pStat->nTotShapeCnt++;              // New # of total Shapes
                        pStat->nTotShapeVertexCnt += nCnt;  // New # of total Shape Vertices
                      }

                    break;

                default:
                    ASSERT(0);                          // Not trapping a Label we spent time looking for!
                    break;
                  }
              }
          }

            //  Beginning of sub block (in one level)

        else if ('{' == *field)
          {
            Node *pN;

                //  If key is unknown, then we don't care about the data!
            pN = (iKEY_UNKNOWN==keyId) ? NULL : pNode;

            if (!aseParseNest(&p, keyId, pN, ma, pStat))
              {
                ret = false;                            // *** Error Abort ***
                break;
              }
          }

            //  End of sub block (out one level)

        else if ('}' == *field)
          {
            ma->nCnt = 0;
            ma->pVertList = NULL;
            break;
          }
      }

pexit:
    *ppStr = p;

    return ret;
}



//
//  ASE parser
//

bool aseParse( const char * pStr, ulong nLen, Node * pNodeBase, rdbInfo *pStat )
{
    char    field[256];
    const char *p;
    MeshArgs    ma;
    
    ASSERT_PTR(pStr);
    ASSERT_ZERO(nLen);
    ASSERT_PTR(pNodeBase);

        //  Integrity checks
#ifdef ASSERT
    aseIntegrityCheck(pStr, nLen);
#endif

    p = skipspace(pStr);
    p = getname( field, p );

        //  Is this a 3DSMAX ASE File?

    if (('*' != *field) || (strcmp( xKEY_3DSMAX_ASCIIEXPORT, field+1)) )
      {
        ELOG( "This is not an 3DSMAX ASE (ASCII Export) Files!\n" );
        return false;
      }

    ramZero(&ma, sizeof(ma));
    pNodeBase->keyId = iKEY_UNKNOWN;

        //  Start parsing with a recursive call!

    if (!aseParseNest( &p, iKEY_UNKNOWN, pNodeBase, &ma, pStat))
      {
        return false;
      }

    if (p != pStr+nLen)
      {
        ELOG( "Nesting Error! (Braces aren't balanced!}\n" );
        ASSERT(0);                          // Brace missing then corrupt!
        return false;
      }

    return true;
}


